from nvdaBuiltin.appModules.tween import AppModule as BuiltinAppModule

# This changes the maximum size of text NVDA retrieves for a list item/column up to 16 K.
# This is the next power of 2 above the maximum length of a DM, which at this writing is 10,000 characters.
# ToDo: This is a global NVDA change from an appModule, which is considered a risky practice, though it seems to work fine in this case.
# [DGL, 2018-03-03]
from NVDAObjects.IAccessible import sysListView32
sysListView32.CBEMAXSTRLEN = max(sysListView32.CBEMAXSTRLEN, 16384)

import api, windowUtils, controlTypes
import ui, scriptHandler
from config import conf
import speech
from NVDAObjects import NVDAObject
from NVDAObjects.window import Window
from . import winforms
import gui
from gui import guiHelper
from globalCommands import commands
from logHandler import log
import winUser, sayAllHandler, queueHandler
import keyboardHandler
import wx, threading, time
from collections import namedtuple


class Tab(namedtuple('Tab', ('name', 'obj', 'index', 'displayName'))):
	"""Describes a single tab in OpenTween.
	Fields description:
		* Name - Name of this tab
		* Obj - NVDA object that represents this tab.
		* Index - The 0-based index of this tab among all active/visible tabs.
		* DisplayName - Speakable/Braillable name used in lists of tabs Etc.
			- (Example: leading "user:" is removed from user-timeline tabs.)
	"""
	pass


class ListDialog(wx.Dialog):
	"""A dialog for selecting a single item among many, consisting of a list control with Ok and Cancel buttons.
	To use:
		- Subclass this base dialog.
		- Optionally Provide a parent, dialog title, and/or label for the list box.
		- Override setItems to preprocess any sequence of items before addition to the list.
			Call the superclass method to populate the list after the necessary filtering.
		Optionally override onOk and onCancel to react to the user pressing the Ok and Cancel buttons.
			If extending, call the superclass methods to hide and destroy the dialog.
	"""

	CENTER_ON_SCREEN = getattr(wx, "CENTER_ON_SCREEN", 2)
	default_parent = gui.mainFrame
	default_title = _("Choose an Item")
	default_list_label = _("Items:")
	def __init__(self, parent=None, title=None, listLabel=None):
		parent = parent if parent is not None else self.__class__.default_parent
		title = title if title is not None else self.__class__.default_title
		lbLabel = listLabel if listLabel is not None else self.__class__.default_list_label
		super(ListDialog, self).__init__(parent, title=title)
		mainSizer = wx.BoxSizer(wx.VERTICAL)
		listSizer = wx.BoxSizer(wx.VERTICAL)
		listHelper = guiHelper.BoxSizerHelper(self, sizer=listSizer)
		listStyle = wx.LC_REPORT | wx.LC_SINGLE_SEL
		self.listBox = listHelper.addLabeledControl(lbLabel, gui.nvdaControls.AutoWidthColumnListCtrl, style=listStyle, id=wx.NewId(), size=(150,250))
		self.listBox.InsertColumn(0, "")
		self.listBox.Bind(wx.EVT_CHAR, self.onListChar)
		mainSizer.Add(listSizer, border=guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
		buttonsSizer = self.CreateButtonSizer(wx.OK|wx.CANCEL)
		mainSizer.Add(buttonsSizer, border=guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL|wx.ALIGN_RIGHT)
		mainSizer.Fit(self)
		self.SetSizer(mainSizer)
		self.Bind(wx.EVT_BUTTON, self.onOk, id=wx.ID_OK)
		self.Bind(wx.EVT_BUTTON, self.onCancel, id=wx.ID_CANCEL)
		self.Center(wx.BOTH | self.CENTER_ON_SCREEN)

	def setItems(self, itemsList):
		"""Populate the list control with any list of items.
		Subclasses should override this method to provide a suitable sequence.
		Call the superclass to perform the actual list population.
		"""
		if not isinstance(itemsList, list): itemsList = list(itemsList)
		for item in itemsList:
			self.listBox.Append((item,))

	def onListChar(self, evt):
		"Here so that pressing Enter in the list activates the default (OK) button."
		#NVDA issue #6206: https://github.com/nvaccess/nvda/issues/6206
		if evt.KeyCode == wx.WXK_RETURN:
			self.ProcessEvent(wx.CommandEvent(wx.wxEVT_COMMAND_BUTTON_CLICKED, wx.ID_OK))
			return
		evt.Skip()

	def onOk(self, evt):
		"""React to the user pressing the Ok button.
		Subclasses can optionally override this method.
		"""
		self.Hide()
		self.Destroy()
	def onCancel(self, evt):
		"""React to the user pressing the Cancel button.
		Subclasses can optionally override this method.
		"""
		self.Hide()
		self.Destroy()


class TabsListDialog(ListDialog):
	default_title = _("Pick a Tab")
	default_list_label = _("Open tabs:")
	def setItems(self, itemsList):
		itemsList.sort(key=lambda tab: tab.displayName.lower())
		self._tabs = itemsList
		super(TabsListDialog, self).setItems((tab.displayName for tab in itemsList))

	def onOk(self, evt):
		super(TabsListDialog, self).onOk(evt)
		api.processPendingEvents()
		tabs = self._tabs
		tabIndex = self.listBox.GetFirstSelected()
		def postProcess():
			speech.cancelSpeech()
			try:
				app = api.getForegroundObject().appModule
				app.switchTab(tabs[tabIndex])
			except Exception as e:
				wx.CallAfter(gui.messageBox, str(e), _("Error Switching Tabs"), wx.OK|wx.ICON_ERROR)
				log.exception(_("Error in postProcess on ok"))
		speech.cancelSpeech()
		wx.CallLater(100,postProcess)


class AppModule(BuiltinAppModule):
	__gestures = {}
	# Actual app name, not made localizeable here.
	scriptCategory = "OpenTween"

	def event_NVDAObject_init(self, obj):
		super(AppModule, self).event_NVDAObject_init(obj)
		try: self._inSpeakForeground
		except AttributeError: self._inSpeakForeground = False
		if not self._inSpeakForeground: return
		role = obj.role
		hwnd = winUser.getForegroundWindow()
		if role == controlTypes.ROLE_WINDOW and hwnd != obj.windowHandle:
			#This avoids double-speaking fields, while also allowing the foreground window to be included in speech generation.
			#ToDO: Could be too permissive
			obj.presentationType = obj.presType_layout
		elif role == controlTypes.ROLE_DOCUMENT:
			#This keeps the first line of document text out of speech generation for any root document objects.
			#sayAllHandler.readObjects generates/speaks full document text while processing each document's subtree.
			#ToDo: This is seemingly private property access (leading underline)
			obj._hasNavigableText = False

	def _firstNamedParent(self, startObj):
		obj = startObj.parent
		while obj:
			if obj.name: break
			obj = obj.parent
		return obj
	def _firstNamedChild(self, startObj):
		obj = startObj.firstChild
		while obj:
			if obj.name: break
			obj = obj.firstChild
		return obj
	def _siblings(self, startObj):
		"Return a generator of objects at the given object's hierarchical level."
		#ToDo: currently not used
		obj = startObj.parent.firstChild
		while obj:
			yield obj
			obj = obj.next

	def findTweetDetailsArea(self):
		"""Find and return the tweet details main area. Raises LookupError if unsuccessful
		"""
		# This is actually the HTML document containing the entire post body.
		fg = api.getForegroundObject().windowHandle
		hwnd = windowUtils.findDescendantWindow(fg, visible=True, className='Internet Explorer_Server')
		return Window(windowHandle=hwnd)

	def findTabControl(self):
		"""Find the containing tab control for tweet lists/buffer tabs.
		Raises LookupError if unsuccessful.
		"""
		#ToDO: Implement caching
		fg = api.getForegroundObject()
		tabWindow = winforms.findWindow(fg, 'SysTabControl32', visible=True)
		#TabWindow is the tab's root window
		area = self._firstNamedParent(tabWindow)
		if isinstance(area, NVDAObject) and area.name == 'PostsListArea':
			tabControl = tabWindow.firstChild
			if (tabControl.role == controlTypes.ROLE_TABCONTROL
					and tabControl.windowClassName == tabWindow.windowClassName
					and tabControl.windowHandle == tabWindow.windowHandle
			):
				return tabControl
		raise LookupError


	__gestures["kb:nvda+shift+d"] = "focusPostDetails"
	def script_focusPostDetails(self, gesture):
		try:
			details = self.findTweetDetailsArea()
			if not isinstance(details, NVDAObject):
				raise LookupError
		except LookupError:
			ui.message("Tweet details area not found")
			return
		details.setFocus()
	script_focusPostDetails.__doc__ = _("Focuses the tweet details area for the selected tweet or DM")

	__gestures["kb:control+shift+d"] = "readPostDetails"
	def script_readPostDetails(self, gesture):
		try:
			details = self.findTweetDetailsArea()
			first = self._firstNamedParent(details)
			if not isinstance(first, NVDAObject):
				raise LookupError
			first = first.parent.parent.firstChild
			textList = []
			#The sender of the tweet (display name/screen name) and any RT information
			textList.append(first.next.name)
			#Text of the post body
			textList.append(details.basicText)
			#tweet timestamp
			textList.append(first.next.next.name)
			#The originating client
			textList.append(first.parent.lastChild.name)
			if scriptHandler.getLastScriptRepeatCount():
				ui.browseableMessage("\n".join(textList))
			else:
				ui.message("; ".join(textList))
		except LookupError:
			ui.message("Tweet details area not found")
	script_readPostDetails.__doc__ = _("Reads the tweet details area for the selected tweet or DM. When pressed twice quickly, places the detail area text in a virtual view for closer examination")

	def event_typedCharacter(self, obj, nextHandler, ch=None):
		"""Squelch Space when used to advance among unread tweets in case speakTypedCharacters is enabled by the user.
		"""
		sect = conf["keyboard"]
		key = "speakTypedCharacters"
		curval = sect[key]
		squelching = (ch == " " and curval == True and self._spaceWillReadTweet(obj))
		if not squelching:
			return nextHandler()
		# Temporarily switch off the speaking of typed characters for this key.
		try:
			sect[key] = False
			nextHandler()
		finally:
			sect[key] = curval

	def _spaceWillReadTweet(self, obj):
		"""Returns True if typing a space will read a tweet rather than doing anything else.
		This should happen exactly if any one of the following is true:
			- Focus is in a tweet buffer (including the DM buffer).
			- Focus is in the posting box and the box is empty.
		"""
		if obj.name == "PostBodyInput": return (not obj.value or len(obj.value) == 0)
		if obj.role not in [controlTypes.ROLE_LISTITEM, controlTypes.ROLE_LIST]: return False
		# FIXME: This is a bit overpermissive.
		return True

	def countTabs(self, tabControlObj):
		"Return the number of open tabs/buffer lists."
		#ToDo: currently used only in testing
		#The even-number check prevents off-by-one because "favorites" tab consistently only seems to spawn a tabControl with no corresponding window.
		childCount = tabControlObj.childCount
		return childCount / 2 + (childCount & 1)

	def _cleanName(self, name):
		"""Clean a tab name for easier sorting in a list and/or displaying to the user."""
		if name.startswith("user:"): name = name[5:]
		return name

	def getTabs(self, tabControlObj=None):
		"""Get and return a list of all currently open tabs/buffers.
		Each tab is represented by a Tab namedtuple. See that class's doc string for record details.
		"""
		if not isinstance(tabControlObj, NVDAObject):
			try: tabControlObj = self.findTabControl()
			except LookupError: pass
		if (not isinstance(tabControlObj, NVDAObject)
				or tabControlObj.role != controlTypes.ROLE_TABCONTROL
				or controlTypes.STATE_INVISIBLE in tabControlObj.states
		):
			return []
		tabsList = []
		for tabIdx, obj in enumerate((ch for ch in tabControlObj.children if ch.role == controlTypes.ROLE_TAB)):
			name = obj.name
			displayName = self._cleanName(name)
			tabsList.append(Tab(name, obj, tabIdx, displayName))
		return tabsList

	def switchTab(self, tab):
		"""Switch to this tab/buffer list.
		"""
		def showErrorWrapper(func, errorMsg, *args, **kwargs):
			speech.cancelSpeech()
			curMode = speech.speechMode
			speech.speechMode = speech.speechMode_off
			try:
				func(*args, **kwargs)
			except Exception as e:
				wx.CallAfter(gui.messageBox, str(e), errorMsg, wx.OK|wx.ICON_ERROR)
				log.exception(_("Error in calling func {0} from switchTabs".format(func.__name__)))
			finally:
				speech.speechMode = curMode
		if controlTypes.STATE_SELECTED in tab.obj.states:
			speech.cancelSpeech()
			queueHandler.queueFunction(queueHandler.eventQueue, ui.message, _("{0} tab already focused".format(tab.displayName)))
			return
		tab.obj.setFocus()
		api.processPendingEvents()
		queueHandler.queueFunction(queueHandler.eventQueue, showErrorWrapper, self.focusPostsList, _("Error Switching Tabs"), tab)

	def focusPostsList(self, tab):
		"""Focus this tab/buffer's post list so that navigation is possible.
		"""
		#ToDo: implement try-n-times behavior if needed
		focus = api.getFocusObject()
		if focus.windowHandle != tab.obj.windowHandle or focus.role != tab.obj.role:
			log.warning(_("Focus: expected {0} tab, got {1} {2}, {3!r}".format(
					tab.displayName, focus.name or 'unnamed', controlTypes.roleLabels[focus.role], focus
			)))
			raise ValueError(_("{0} tab not focused".format(tab.displayName)))
		try: keyboardHandler.KeyboardInputGesture.fromName("tab").send()
		except Exception: raise ValueError(_("{0} tab: Posts list not focused.".format(tab.displayName)))

	def showDialog(self, dlg):
		if not dlg.IsShown():
			gui.mainFrame.prePopup()
			dlg.Show()
			gui.mainFrame.postPopup()

	__gestures["kb:control+shift+t"] = "pickATab"
	def script_pickATab(self, gesture):
		try:
			tabs = self.getTabs()
			dlg = TabsListDialog()
			dlg.setItems(tabs)
			dlg.listBox.SetFocus()
			self.showDialog(dlg)
		except Exception as e:
			log.exception(_("Error in `pick a tab' script:"))
			wx.CallAfter(gui.messageBox, _("Cannot gather or present tabs"), _("ERROR"), wx.OK|wx.ICON_ERROR)
	script_pickATab.__doc__ = _("Select the tab to switch to from a list of active tabs")

	def _speakForegroundWatcher(self):
		"""Monitor for stoppage  of a sayAll that was innitiated by speakForeground script (default gesture Nvda+b ).
		toggles a flag used to filter extraneous windows from foreground object reading.
		"""
		#Called from a separate thread.
		while sayAllHandler.isRunning():
			time.sleep(0.01)
		self._inSpeakForeground = False

	__gestures["kb:nvda+b"] = "openTweenSpeakForeground"
	def script_openTweenSpeakForeground(self, gesture):
		"""Global override to silence double-speaking of fields in some OpenTween dialogs.
		"""
		self._inSpeakForeground = True
		th = threading.Thread(target=self._speakForegroundWatcher)
		commands.script_speakForeground(None)
		queueHandler.queueFunction(queueHandler.eventQueue, th.start)

	__gestures["kb:nvda+shift+v"] = "announceAppVersion"
	def script_announceAppVersion(self, gesture):
		try:
			ui.message("{0} version {1}".format(self.productName, self.productVersion))
		except:
			ui.message(_("Version information not available"))
	script_announceAppVersion.__doc__ = _("Announce app name and version")
