#Collection of utilities for working with Winforms components

#ToDo:
	#- Implement caching
	#- Raise custom exception when a valid-looking Winforms class cannot be constructed


import windowUtils
from NVDAObjects.window import Window


PART_SEP = '.'
#Parts list is in order of appearance in a Winforms class string
PARTS_LIST = ['prefix', 'className', 'style', 'suffix']

class WinformsClass(object):
	"Represents a Winforms class and allows access to its various component parts"

	def __init__(self, prefix, className, suffix, style=None):
		self.prefix = prefix
		self.className = className
		self.suffix = suffix
		if not style: style = ''
		self.style = str(style)

	def _isBlank(self, s):
		if len(s) == 0 or s.isspace():
			return True
		return False

	def fullName(self, includeStyle=False):
		if includeStyle and not self.style: includeStyle = False
		name = ''
		textList = []
		for part in PARTS_LIST:
			if part == 'style' and not includeStyle: continue
			attr = getattr(self, part, None)
			if attr and isinstance(attr, str) and not self._isBlank(attr):
				textList.append(attr)
		targetLen = len(PARTS_LIST)
		textlen = len(textList) 
		if (includeStyle and textlen == targetLen
				or (not includeStyle and textlen == targetLen - 1)
		):
			name = PART_SEP.join(textList)
		return name

	@classmethod
	def fromObj(cls, obj):
		"Return a Winforms class constructed from the given object's window class name."
		wclass = ''
		try: wclass = obj.windowClassName
		except Exception: pass
		if not wclass or PART_SEP not in wclass: return ''
		suffixIdx = wclass.lower().rfind('{0}{1}{0}'.format(PART_SEP, 'app', PART_SEP))
		if suffixIdx < 0: return ''
		suffix = wclass[suffixIdx + 1:]
		wclass = wclass[:suffixIdx]
		classParts = wclass.split(PART_SEP)
		prefix = classParts[0]
		className = classParts[1]
		style = None
		try:
			style = int(classParts[2])
		except Exception:
			pass
		return cls(prefix, className, suffix,  style)

def findWindow(parent, className, visible=None, controlID=None):
	"""Find a descendant window of the given parent NVDA object by constructing the Winforms class name from the given (normalized) window class.
	Other arguments are as for windowUtils.findDescendantWindow(). Returns an appropriate NVDA object if found  and raises LookupError otherwise.
	"""
	parentHandle = parent.windowHandle
	parentClass = WinformsClass.fromObj(parent)
	if not parentClass.prefix.lower().startswith("windowsforms"): raise LookupError
	targetClass = WinformsClass(parentClass.prefix, className, parentClass.suffix, parentClass.style)
	#ToDo: May also be other classes that seem to include style.
	includeStyle = True if className.lower() == 'window' else False
	tries = 2
	while tries > 0:
		try:
			hwnd = windowUtils.findDescendantWindow(parentHandle, visible=visible, controlID=controlID, className=targetClass.fullName(includeStyle))
			kwargs = { 'windowHandle': hwnd}
			cls = Window.findBestAPIClass(kwargs)
			return cls(**kwargs)
		except LookupError:
			pass
		tries -= 1
		includeStyle = not includeStyle
	raise LookupError
